import { html } from "uhtml";
import { signal } from "uhtml/preactive";
import { API_BASE_URL } from "../../utils/store.ts";
import { retrieveOrGenerateKeyPair } from "../../utils/crypto.ts";
import { openSettings } from "./Settings.ts";

const notificationPermission = signal(Notification.permission);
const pushSubscription = signal<PushSubscription | null>(null);
export const NotificationManager = () => html` <h2>Notifications</h2>
  <p>
    This is an <mark>experimental feature</mark> and may not work as expected.
    It will not work in offline-mode, and requires subscribing to the VAPID Web
    Push server at <code>${API_BASE_URL}</code> for it to send Push
    Notifications to your device in the background.
  </p>
  <p>
    If enabled, you will receive a notification whenever a user generates a URL
    with a message to send to you. Upon clicking on the notification, it will
    immediately take you to their generated URL and reveal the message.
  </p>
  <p>Notifications: <code>${notificationPermission.value}</code></p>
  <p>Subscribed: <code>${pushSubscription.value !== null}</code></p>
  <p>
    <button
      @click=${subscribeUserToPush}
      ?disabled=${notificationPermission.value === "granted"}
    >
      Enable Notifications
    </button>
    <button
      @click=${unsubscribeUserFromPush}
      ?disabled=${pushSubscription.value === null}
    >
      Unsubscribe
    </button>
    <button class="secondary" @click=${openSettings}>Back</button>
  </p>`;

export async function sendNotification(publicKey: CryptoKey, url: URL) {
  const exportedKey = await window.crypto.subtle.exportKey("raw", publicKey);
  const publicKeyString = btoa(
    String.fromCharCode(...new Uint8Array(exportedKey))
  );

  await fetch(`${API_BASE_URL}api/send-notification`, {
    method: "POST",
    body: JSON.stringify({
      publicKey: publicKeyString,
      url: url.toString(),
    }),
    headers: {
      "Content-Type": "application/json",
    },
  });
}

export async function checkPushSubscription() {
  const registration = await navigator.serviceWorker.ready;
  const subscription = await registration.pushManager.getSubscription();
  pushSubscription.value = subscription;
}

export async function unsubscribeUserFromPush() {
  try {
    const registration = await navigator.serviceWorker.ready;
    const subscription = await registration.pushManager.getSubscription();

    if (!subscription) {
      throw new Error("No subscription to unsubscribe");
    }

    const successful = await subscription.unsubscribe();
    if (!successful) {
      throw new Error("Failed to unsubscribe");
    }

    // await fetch(`${API_BASE_URL}api/unsubscribe`, {
    //   method: "POST",
    //   headers: {
    //     "Content-Type": "application/json",
    //   },
    //   body: JSON.stringify({ subscription }),
    // });

    pushSubscription.value = null;
    console.log("User is unsubscribed.");
  } catch (err) {
    console.error("Failed to unsubscribe the user: ", err);
  }
}

export async function subscribeUserToPush(
  event: Event | undefined = undefined
) {
  event?.preventDefault();
  try {
    const permission = await Notification.requestPermission();
    if (permission !== "granted") {
      throw new Error("Notification permission not granted");
    }

    notificationPermission.value = permission;

    const registration = await navigator.serviceWorker.ready;
    let subscription = await registration.pushManager.getSubscription();
    const response = await fetch(`${API_BASE_URL}api/public-key`);
    const { publicKey: serverPublicKey } = await response.json();
    if (!subscription) {
      subscription = await registration.pushManager.subscribe({
        userVisibleOnly: true,
        applicationServerKey: serverPublicKey,
      });
    }

    const { publicKey: publicKey } = await retrieveOrGenerateKeyPair();

    const exportedKey = await window.crypto.subtle.exportKey("raw", publicKey);
    const exportedKeyString = btoa(
      String.fromCharCode(...new Uint8Array(exportedKey))
    );

    await fetch(`${API_BASE_URL}api/subscribe`, {
      method: "POST",
      headers: {
        "Content-Type": "application/json",
      },
      body: JSON.stringify({ subscription, publicKey: exportedKeyString }),
    });

    console.log("User is subscribed.");
    checkPushSubscription();
  } catch (err) {
    console.error("Failed to subscribe the user: ", err);
  }
}
