import { LOCAL_STORAGE_KEYS } from "./store";

export async function generateKeyPair(): Promise<CryptoKeyPair> {
  return await window.crypto.subtle.generateKey(
    {
      name: "ECDH",
      namedCurve: "P-256",
    },
    true,
    ["deriveKey", "deriveBits"]
  );
}

export async function deriveSharedSecret(
  privateKey: CryptoKey,
  publicKey: CryptoKey
): Promise<CryptoKey> {
  return await window.crypto.subtle.deriveKey(
    {
      name: "ECDH",

      public: publicKey,
    },
    privateKey,
    {
      name: "AES-GCM",
      length: 256,
    },
    true,
    ["encrypt", "decrypt"]
  );
}

export async function retrieveOrGenerateKeyPair(): Promise<CryptoKeyPair> {
  let ecdhPublic: CryptoKey;
  let ecdhPrivate: CryptoKey;

  if (
    localStorage.getItem(LOCAL_STORAGE_KEYS.ECDH_PUBLIC_KEY) &&
    localStorage.getItem(LOCAL_STORAGE_KEYS.ECDH_PRIVATE_KEY)
  ) {
    const ecdhPublicJwk = JSON.parse(
      localStorage.getItem(LOCAL_STORAGE_KEYS.ECDH_PUBLIC_KEY)!
    );
    const ecdhPrivateJwk = JSON.parse(
      localStorage.getItem(LOCAL_STORAGE_KEYS.ECDH_PRIVATE_KEY)!
    );

    ecdhPrivate = await window.crypto.subtle.importKey(
      "jwk",
      ecdhPrivateJwk,
      {
        name: "ECDH",
        namedCurve: "P-256",
      },
      true,
      ["deriveKey", "deriveBits"]
    );

    ecdhPublic = await window.crypto.subtle.importKey(
      "jwk",
      ecdhPublicJwk,
      {
        name: "ECDH",
        namedCurve: "P-256",
      },
      true,
      []
    );
  } else {
    const keyPair = await generateKeyPair();
    ecdhPublic = keyPair.publicKey;
    ecdhPrivate = keyPair.privateKey;
    saveKeys(ecdhPublic, ecdhPrivate);
  }
  return { publicKey: ecdhPublic, privateKey: ecdhPrivate };
}

async function saveKeys(ecdhPublic: CryptoKey, ecdhPrivate: CryptoKey) {
  const ecdhPublicJwk = await window.crypto.subtle.exportKey("jwk", ecdhPublic);
  const ecdhPrivateJwk = await window.crypto.subtle.exportKey(
    "jwk",
    ecdhPrivate
  );

  // Store ECDH key pair in local storage
  localStorage.setItem(
    LOCAL_STORAGE_KEYS.ECDH_PUBLIC_KEY,
    JSON.stringify(ecdhPublicJwk)
  );
  localStorage.setItem(
    LOCAL_STORAGE_KEYS.ECDH_PRIVATE_KEY,
    JSON.stringify(ecdhPrivateJwk)
  );

  return {
    public: JSON.stringify(ecdhPublicJwk),
    private: JSON.stringify(ecdhPrivateJwk),
  };
}

export async function encrypt(
  aesKey: CryptoKey,
  data: string
): Promise<{ encryptedData: ArrayBuffer; iv: Uint8Array }> {
  // Convert the data to a Uint8Array
  const dataUint8Array = new TextEncoder().encode(data);

  // Generate a random initialization vector (iv)
  const iv = window.crypto.getRandomValues(new Uint8Array(12));

  // Encrypt the data
  const encryptedData = await window.crypto.subtle.encrypt(
    {
      name: "AES-GCM",
      iv: iv,
    },
    aesKey,
    dataUint8Array
  );

  // Return the combined iv and encrypted data
  return { encryptedData, iv };
}

export async function decrypt(
  aesKey: CryptoKey,
  iv: string,
  encryptedData: string
): Promise<string> {
  // Decode the iv and encryptedData from base64
  const ivUint8Array = base64ToUint8Array(iv);
  const encryptedDataUint8Array = base64ToUint8Array(encryptedData);

  // Decrypt the data
  const decryptedData = await window.crypto.subtle.decrypt(
    {
      name: "AES-GCM",
      iv: ivUint8Array,
    },
    aesKey,
    encryptedDataUint8Array
  );

  // Convert the decrypted data from a Uint8Array to a string
  const decryptedString = new TextDecoder().decode(decryptedData);

  return decryptedString;
}

function base64ToUint8Array(base64: string): Uint8Array {
  const binaryString = atob(base64);
  const len = binaryString.length;
  const bytes = new Uint8Array(len);
  for (let i = 0; i < len; i++) {
    bytes[i] = binaryString.charCodeAt(i);
  }
  return bytes;
}

function downloadFile(content: string, fileName: string, contentType: string) {
  const a = document.createElement("a");
  const file = new Blob([content], { type: contentType });

  a.href = URL.createObjectURL(file);
  a.download = fileName;
  a.click();

  URL.revokeObjectURL(a.href);
}

export async function exportKeys(): Promise<void> {
  const keys = await retrieveOrGenerateKeyPair();

  const exportedKeys = await saveKeys(keys.publicKey, keys.privateKey);
  const exportedKeysJson = JSON.stringify(exportedKeys);

  downloadFile(exportedKeysJson, "keys.json", "application/json");
}

export async function importAndSaveKeys(keys: {
  public: string;
  private: string;
}): Promise<void> {
  const ecdhPublicJwk = JSON.parse(keys.public);
  const ecdhPrivateJwk = JSON.parse(keys.private);

  // Import the keys
  const ecdhPublic = await window.crypto.subtle.importKey(
    "jwk",
    ecdhPublicJwk,
    {
      name: "ECDH",
      namedCurve: "P-256",
    },
    true,
    []
  );

  const ecdhPrivate = await window.crypto.subtle.importKey(
    "jwk",
    ecdhPrivateJwk,
    {
      name: "ECDH",
      namedCurve: "P-256",
    },
    true,
    ["deriveKey", "deriveBits"]
  );

  // Save the imported keys
  saveKeys(ecdhPublic, ecdhPrivate);
}
